#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// maximum length of one word, in letters
#define MAX_WORD_LENGTH 20

// We have a file in the following format:
//
// 3
// Apple
// Giraffe
// Hover
//
// ...where the first line gives the number of words in the
// file, and every subsequent line is a word.  Read this into
// an array of type char**.  The name of the file is provided on
// the command line.

int main(int argc, char** argv) {
  if (argc != 2) {
    // first argument: name of program being run
    // second argument: name of file to open
    printf("Needs a filename\n");
    return 1; // non-zero exit code indicates failure
  }

  // if we get here, we have a file name to open

  FILE* file = fopen(argv[1], "r"); // open file for reading
  if (file == NULL) {
    printf("Could not open file '%s'\n", argv[1]);
    return 2;
  }

  // if we get here, we were able to open the file

  // read in the number of words in the file
  int numWords;
  fscanf(file, "%d", &numWords);

  if (numWords < 0) {
    printf("Needs a non-negative number of words; received %i\n", numWords);
    fclose(file); // be sure to close the file
    return 3;
  } else if (numWords == 0) {
    // The file contains no words, so it is pointless to move forward
    fclose(file);
    return 0; // not an error condition, so return 0
  }

  // If we get here, we have a positive number of words to allocate.
  // Before we can allocate, we must read a bit more from the file.
  // The number should be followed by a newline, and we must read past
  // this point for subsequent reading to work correctly.
  int c = fgetc(file);
  
  if (c == EOF) {
    printf("WARNING: File declares a non-negative number of words to");
    printf(" read in, but the file contains no words!\n");
    fclose(file);
    return 3;
  } else if ((char)c != '\n') {
    printf("ERROR: The number of words is not followed by a newline\n");
    fclose(file);
    return 4;
  }

  // If we get here, we are ready to start reading in words.
  // Allocate the space for the pointers to words.

  char** words = (char**)malloc(sizeof(char*) * numWords);

  if (words == NULL) {
    printf("Failed to allocate space for all words!\n");
    fclose(file);
    return 5;
  }

  // Allocate buffer for reading in the current word.  Basic idea
  // is to read into the buffer, and then copy the content from the
  // buffer into a new space which is exactly as large as necessary
  // for the new word.  We could just always allocate the maximum
  // size, but this is wasteful in the case of smaller words.
  char buffer[MAX_WORD_LENGTH + 2]; // +1 for terminating null byte and newline
  int currentWord = 0;

  while (currentWord < numWords && fgets(buffer, MAX_WORD_LENGTH + 2, file)) {
    // fgets will include an unwanted newline.  If the last letter in
    // the word is a newline, we want to zero it out with a null byte
    int len = strlen(buffer);
    if (buffer[len - 1] == '\n') {
      len--;
      buffer[len] = '\0';
    }

    // allocate space for this word.  +1 for null byte
    words[currentWord] = (char*)malloc(sizeof(char) * (len + 1));

    if (words[currentWord] == NULL) {
      printf("Failed to allocate space for a word!\n");
      fclose(file);

      // we should deallocate the words we've already allocated here,
      // but the operating system will take care of this for us on exit
      // anyway

      return 6;
    }

    // copy the word over
    strcpy(words[currentWord], buffer);

    // onto the next word
    currentWord++;
  }

  // At this point, we should have read in the words.  However, there
  // are two possible warning/error conditions:
  //
  // 1. The number of words declared was too many.  If this is true,
  //    we must have read in fewer than `numWords` words.  This wastes
  //    memory, since we allocated space for more words than necessary.
  //
  // 2. The number of words declared was too few.  If this is true, then
  //    we are not at the end of the file.  This prevents all words from
  //    being read in

  if (currentWord < numWords) {
    printf("WARNING: Memory wasted - File declared it contained");
    printf(" %i words but only held %i words.\n", numWords, currentWord);
  } else if (!feof(file)) { // not at end of file
    printf("WARNING: Not all words read in - The file contains more");
    printf(" than the %i words declared.\n", numWords);
  }

  // close the file
  fclose(file);

  // print out the words
  int x;
  for(x = 0; x < currentWord; x++) {
    // NOT numWords - currentWord holds the number of words _actually_
    // read in, whereas numWords holds the number of words expected
    printf("%s\n", words[x]);
  }

  // deallocate the individual words
  for(x = 0; x < currentWord; x++) {
    free(words[x]);
  }

  // free the space for all the pointers
  free(words);

  // indicate success and terminate
  return 0;
}

    
